@@ -1,4 +1,4 @@
-/* $Id: shutdown.c 4713 2004-11-02 16:43:43Z benny $ */
+/* $Id: shutdown.c 29056 2008-12-27 04:59:51Z kelnos $ */
 /*-
  * Copyright (c) 2003-2004 Benedikt Meurer <benny@xfce.org>
  * All rights reserved.
@@ -40,6 +40,19 @@
 #include <unistd.h>
 #endif
 
+#ifdef HAVE_GETPWUID
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
+#ifdef HAVE_PWD_H
+#include <pwd.h>
+#endif
+#endif
+
+#ifdef HAVE_ASM_UNISTD_H
+#include <asm/unistd.h>  /* for __NR_ioprio_set */
+#endif
+
 #include <libxfce4util/libxfce4util.h>
 #include <libxfcegui4/libxfcegui4.h>
 #include <gtk/gtk.h>
@@ -54,15 +67,16 @@
 #include <xfce4-session/xfsm-legacy.h>
 #include <xfce4-session/xfsm-shutdown-helper.h>
 
-#define BORDER		6
+#define BORDER    6
 
 
 static XfsmShutdownHelper *shutdown_helper = NULL;
 
+static GtkWidget *shutdown_dialog = NULL;
 
 #ifdef SESSION_SCREENSHOTS
 static void
-screenshot_save (GdkPixmap *pm, GdkRectangle *area)
+screenshot_save (const gchar *session_name, GdkPixmap *pm, GdkRectangle *area)
 {
   gchar *display_name;
   gchar *resource;
@@ -109,49 +123,96 @@ entry_activate_cb (GtkWidget *entry, GtkDialog *dialog)
   gtk_dialog_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
 }
 
+static void
+logout_button_clicked (GtkWidget *b, gint *shutdownType)
+{
+    *shutdownType = XFSM_SHUTDOWN_LOGOUT;
+
+    gtk_dialog_response (GTK_DIALOG (shutdown_dialog), GTK_RESPONSE_OK);
+}
+
+static void
+reboot_button_clicked (GtkWidget *b, gint *shutdownType)
+{
+    *shutdownType = XFSM_SHUTDOWN_REBOOT;
+
+    gtk_dialog_response (GTK_DIALOG (shutdown_dialog), GTK_RESPONSE_OK);
+}
+
+static void
+halt_button_clicked (GtkWidget *b, gint *shutdownType)
+{
+    *shutdownType = XFSM_SHUTDOWN_HALT;
+
+    gtk_dialog_response (GTK_DIALOG (shutdown_dialog), GTK_RESPONSE_OK);
+}
+
+static void
+suspend_button_clicked (GtkWidget *b, gint *shutdownType)
+{
+    *shutdownType = XFSM_SHUTDOWN_SUSPEND;
+
+    gtk_dialog_response (GTK_DIALOG (shutdown_dialog), GTK_RESPONSE_OK);
+}
+
+static void
+hibernate_button_clicked (GtkWidget *b, gint *shutdownType)
+{
+    *shutdownType = XFSM_SHUTDOWN_HIBERNATE;
+
+    gtk_dialog_response (GTK_DIALOG (shutdown_dialog), GTK_RESPONSE_OK);
+}
 
 /*
  */
 gboolean
-shutdownDialog(gint *shutdownType, gboolean *saveSession)
+shutdownDialog(const gchar *sessionName, XfsmShutdownType *shutdownType, gboolean *saveSession)
 {
-	gboolean accessibility;
+  gboolean accessibility;
   XfsmFadeout *fadeout = NULL;
   GdkScreen *screen;
-	GtkWidget *dialog;
-	GtkWidget *label;
-	GtkWidget *dbox;
-	GtkWidget *hbox;
-	GtkWidget *vbox;
-	GtkWidget *image;
-  GtkWidget *radio_vbox;
-  GtkWidget *radio_logout;
-  GtkWidget *radio_reboot;
-  GtkWidget *radio_halt;
-	GtkWidget *checkbox;
+  GtkWidget *dialog;
+  GtkWidget *label;
+  GtkWidget *dbox;
+  GtkWidget *hbox;
+  GtkWidget *vbox;
+  GtkWidget *vbox2;
+  GtkWidget *image;
+  GtkWidget *checkbox;
   GtkWidget *entry_vbox;
   GtkWidget *entry;
-	GtkWidget *hidden;
-  GtkWidget *ok_button;
+  GtkWidget *hidden;
+  GtkWidget *logout_button;
+  GtkWidget *reboot_button;
+  GtkWidget *halt_button;
+  GtkWidget *suspend_button = NULL;
+  GtkWidget *hibernate_button = NULL;
   GtkWidget *cancel_button;
+  GtkWidget *ok_button;
   GdkPixbuf *icon;
   gboolean saveonexit;
   gboolean autosave;
   gboolean prompt;
+  gboolean show_suspend;
+  gboolean show_hibernate;
   gint monitor;
-	gint result;
+  gint result;
   XfceKiosk *kiosk;
   gboolean kiosk_can_shutdown;
-  XfceRc *rc;
+  gboolean kiosk_can_save_session;
+  XfconfChannel *channel;
 #ifdef SESSION_SCREENSHOTS
   GdkRectangle screenshot_area;
   GdkWindow *root;
   GdkPixmap *screenshot_pm = NULL;
   GdkGC *screenshot_gc;
 #endif
+#ifdef HAVE_GETPWUID
+  struct passwd *pw;
+#endif
 
-	g_return_val_if_fail(saveSession != NULL, FALSE);
-	g_return_val_if_fail(shutdownType != NULL, FALSE);
+  g_return_val_if_fail(saveSession != NULL, FALSE);
+  g_return_val_if_fail(shutdownType != NULL, FALSE);
 
   /* destroy any previously running shutdown helper first */
   if (shutdown_helper != NULL)
@@ -163,21 +224,29 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
   /* load kiosk settings */
   kiosk = xfce_kiosk_new ("xfce4-session");
   kiosk_can_shutdown = xfce_kiosk_query (kiosk, "Shutdown");
+  kiosk_can_save_session = xfce_kiosk_query (kiosk, "SaveSession");
   xfce_kiosk_free (kiosk);
 
   /* load configuration */
-  rc = xfsm_open_config (FALSE);
-  xfce_rc_set_group (rc, "General");
-  saveonexit = xfce_rc_read_bool_entry (rc, "SaveOnExit", TRUE);
-  autosave = xfce_rc_read_bool_entry (rc, "AutoSave", FALSE);
-  prompt = xfce_rc_read_bool_entry (rc, "PromptOnLogout", TRUE);
+  channel = xfsm_open_config ();
+  channel = xfconf_channel_get ("xfce4-session");
+  saveonexit = xfconf_channel_get_bool (channel, "/general/SaveOnExit", TRUE);
+  autosave = xfconf_channel_get_bool (channel, "/general/AutoSave", FALSE);
+  prompt = xfconf_channel_get_bool (channel, "/general/PromptOnLogout", TRUE);
+  show_suspend = xfconf_channel_get_bool (channel, "/shutdown/ShowSuspend", TRUE);
+  show_hibernate = xfconf_channel_get_bool (channel, "/shutdown/ShowHibernate", TRUE);
+
+  /* make the session-save settings obey the kiosk settings */
+  if (!kiosk_can_save_session)
+    {
+      saveonexit = FALSE;
+      autosave = FALSE;
+    }
 
   /* if PromptOnLogout is off, saving depends on AutoSave */
   if (!prompt)
     {
-      xfce_rc_close (rc);
-
-      *shutdownType = SHUTDOWN_LOGOUT;
+      *shutdownType = XFSM_SHUTDOWN_LOGOUT;
       *saveSession = autosave;
 
       return TRUE;
@@ -198,17 +267,17 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
       monitor = 0;
     }
 
-	/* Try to grab Input on a hidden window first */
-	hidden = gtk_invisible_new_for_screen (screen);
-	gtk_widget_show_now (hidden);
+  /* Try to grab Input on a hidden window first */
+  hidden = gtk_invisible_new_for_screen (screen);
+  gtk_widget_show_now (hidden);
 
-	accessibility = GTK_IS_ACCESSIBLE (gtk_widget_get_accessible (hidden));
+  accessibility = GTK_IS_ACCESSIBLE (gtk_widget_get_accessible (hidden));
 
   if (!accessibility) 
     {
       for (;;)
         {
-          if (gdk_pointer_grab (hidden->window, FALSE, 0, NULL, NULL,
+          if (gdk_pointer_grab (hidden->window, TRUE, 0, NULL, NULL,
                                 GDK_CURRENT_TIME) == GDK_GRAB_SUCCESS)
             {
               if (gdk_keyboard_grab (hidden->window, FALSE, GDK_CURRENT_TIME)
@@ -262,57 +331,192 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
       gtk_window_set_decorated (GTK_WINDOW (dialog), FALSE);
     }
 
+  shutdown_dialog = dialog;
+
   cancel_button = gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_CANCEL,
                                          GTK_RESPONSE_CANCEL);
+
   ok_button = gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_OK,
                                      GTK_RESPONSE_OK);
 
-	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL);
+  gtk_widget_hide (ok_button);
+
+  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL);
   gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
-	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
+  gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
   gtk_window_set_screen (GTK_WINDOW (dialog), screen);
 
-	dbox = GTK_DIALOG(dialog)->vbox;
+  dbox = GTK_DIALOG(dialog)->vbox;
 
-	hbox = gtk_hbox_new(FALSE, 0);
-	gtk_box_pack_start(GTK_BOX(dbox), hbox, TRUE, TRUE, BORDER);
-	gtk_widget_show(hbox);
+  vbox = gtk_vbox_new(FALSE, BORDER);
+  gtk_box_pack_start(GTK_BOX(dbox), vbox, TRUE, TRUE, 0);
+  gtk_container_set_border_width (GTK_CONTAINER (vbox), BORDER);
+  gtk_widget_show(vbox);
 
-  icon = xfce_themed_icon_load ("xfsm-shutdown", 48);
-	image = gtk_image_new_from_pixbuf (icon);
-	gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, TRUE, BORDER);
-	gtk_widget_show(image);
+#ifdef HAVE_GETPWUID
+  pw = getpwuid (getuid ());
+  if (G_LIKELY(pw && pw->pw_name && *pw->pw_name))
+    {
+      gchar *text = g_strdup_printf (_("<span size='large'><b>Log out %s</b></span>"), pw->pw_name);
+      GtkWidget *label = g_object_new (GTK_TYPE_LABEL,
+                                       "label", text,
+                                       "use-markup", TRUE,
+                                       "justify", GTK_JUSTIFY_CENTER,
+                                       "xalign", 0.5,
+                                       "yalign", 0.5,
+                                       NULL);
+
+      gtk_widget_show (label);
+      gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
+
+      g_free (text);
+    }
+#endif
+
+  hbox = gtk_hbox_new (TRUE, BORDER);
+  gtk_widget_show (hbox);
+  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
+  
+  /* logout */
+  logout_button = gtk_button_new ();
+  gtk_widget_show (logout_button);
+  gtk_box_pack_start (GTK_BOX (hbox), logout_button, TRUE, TRUE, 0);
+
+  g_signal_connect (logout_button, "clicked", 
+                    G_CALLBACK (logout_button_clicked), shutdownType);
+
+  vbox2 = gtk_vbox_new (FALSE, BORDER);
+  gtk_container_set_border_width (GTK_CONTAINER (vbox2), BORDER);
+  gtk_widget_show (vbox2);
+  gtk_container_add (GTK_CONTAINER (logout_button), vbox2);
+
+  icon = xfce_themed_icon_load ("system-log-out", 32);
+  if (!icon)
+    icon = xfce_themed_icon_load ("xfsm-logout", 32);
+  image = gtk_image_new_from_pixbuf (icon);
+  gtk_widget_show (image);
+  gtk_box_pack_start (GTK_BOX (vbox2), image, FALSE, FALSE, 0);
   g_object_unref (icon);
 
-	vbox = gtk_vbox_new(FALSE, 0);
-	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, BORDER);
-	gtk_widget_show(vbox);
+  label = gtk_label_new (_("Log Out"));
+  gtk_widget_show (label);
+  gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, FALSE, 0);
+  
+  /* reboot */
+  reboot_button = gtk_button_new ();
+  gtk_widget_show (reboot_button);
+  gtk_box_pack_start (GTK_BOX (hbox), reboot_button, TRUE, TRUE, 0);
 
-	label = gtk_label_new(_("What do you want to do next?"));
-	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, TRUE, BORDER);
-	gtk_widget_show(label);
+  g_signal_connect (reboot_button, "clicked", 
+                    G_CALLBACK (reboot_button_clicked), shutdownType);
 
-  radio_vbox = gtk_vbox_new (FALSE, 0);
-  gtk_box_pack_start (GTK_BOX (vbox), radio_vbox, TRUE, FALSE, BORDER);
-  gtk_widget_show (radio_vbox);
+  vbox2 = gtk_vbox_new (FALSE, BORDER);
+  gtk_container_set_border_width (GTK_CONTAINER (vbox2), BORDER);
+  gtk_widget_show (vbox2);
+  gtk_container_add (GTK_CONTAINER (reboot_button), vbox2);
 
-  radio_logout = gtk_radio_button_new_with_label (
-      NULL, _("Quit current session"));
-  gtk_box_pack_start (GTK_BOX (radio_vbox), radio_logout, FALSE, FALSE, 0);
-  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio_logout), TRUE);
-  gtk_widget_show (radio_logout);
+  icon = xfce_themed_icon_load ("system-reboot", 32);
+  if (!icon)
+    icon = xfce_themed_icon_load ("xfsm-reboot", 32);
+  image = gtk_image_new_from_pixbuf (icon);
+  gtk_widget_show (image);
+  gtk_box_pack_start (GTK_BOX (vbox2), image, FALSE, FALSE, 0);
+  g_object_unref (icon);
 
-  radio_reboot = gtk_radio_button_new_with_label_from_widget (
-      GTK_RADIO_BUTTON (radio_logout), _("Reboot the computer"));
-  gtk_box_pack_start (GTK_BOX (radio_vbox), radio_reboot, FALSE, FALSE, 0);
-  gtk_widget_show (radio_reboot);
+  label = gtk_label_new (_("Restart"));
+  gtk_widget_show (label);
+  gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, FALSE, 0);
+  
+  /* halt */
+  halt_button = gtk_button_new ();
+  gtk_widget_show (halt_button);
+  gtk_box_pack_start (GTK_BOX (hbox), halt_button, TRUE, TRUE, 0);
 
-  radio_halt = gtk_radio_button_new_with_label_from_widget (
-      GTK_RADIO_BUTTON (radio_logout), _("Turn off the computer"));
-  gtk_box_pack_start (GTK_BOX (radio_vbox), radio_halt, FALSE, FALSE, 0);
-  gtk_widget_show (radio_halt);
+  g_signal_connect (halt_button, "clicked", 
+                    G_CALLBACK (halt_button_clicked), shutdownType);
 
-  if (!autosave)
+  vbox2 = gtk_vbox_new (FALSE, BORDER);
+  gtk_container_set_border_width (GTK_CONTAINER (vbox2), BORDER);
+  gtk_widget_show (vbox2);
+  gtk_container_add (GTK_CONTAINER (halt_button), vbox2);
+
+  icon = xfce_themed_icon_load ("system-shutdown", 32);
+  if (!icon)
+    icon = xfce_themed_icon_load ("xfsm-shutdown", 32);
+  image = gtk_image_new_from_pixbuf (icon);
+  gtk_widget_show (image);
+  gtk_box_pack_start (GTK_BOX (vbox2), image, FALSE, FALSE, 0);
+  g_object_unref (icon);
+  
+  label = gtk_label_new (_("Shut Down"));
+  gtk_widget_show (label);
+  gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, FALSE, 0);
+  
+  if (show_suspend || show_hibernate)
+    {
+      hbox = gtk_hbox_new (FALSE, BORDER);
+      gtk_widget_show (hbox);
+      gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
+    }
+
+  /* suspend */
+  if (show_suspend)
+    {
+      suspend_button = gtk_button_new ();
+      gtk_widget_show (suspend_button);
+      gtk_box_pack_start (GTK_BOX (hbox), suspend_button, TRUE, TRUE, 0);
+
+      g_signal_connect (suspend_button, "clicked",
+                        G_CALLBACK (suspend_button_clicked), shutdownType);
+
+      vbox2 = gtk_vbox_new (FALSE, BORDER);
+      gtk_container_set_border_width (GTK_CONTAINER (vbox2), BORDER);
+      gtk_widget_show (vbox2);
+      gtk_container_add (GTK_CONTAINER (suspend_button), vbox2);
+
+      icon = xfce_themed_icon_load ("system-suspend", 32);
+      if (!icon)
+        icon = xfce_themed_icon_load ("xfsm-suspend", 32);
+      image = gtk_image_new_from_pixbuf (icon);
+      gtk_widget_show (image);
+      gtk_box_pack_start (GTK_BOX (vbox2), image, FALSE, FALSE, 0);
+      g_object_unref (icon);
+
+      label = gtk_label_new (_("Suspend"));
+      gtk_widget_show (label);
+      gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, FALSE, 0);
+    }
+
+  /* hibernate */
+  if (show_hibernate)
+    {
+      hibernate_button = gtk_button_new ();
+      gtk_widget_show (hibernate_button);
+      gtk_box_pack_start (GTK_BOX (hbox), hibernate_button, TRUE, TRUE, 0);
+
+      g_signal_connect (hibernate_button, "clicked",
+                        G_CALLBACK (hibernate_button_clicked), shutdownType);
+
+      vbox2 = gtk_vbox_new (FALSE, BORDER);
+      gtk_container_set_border_width (GTK_CONTAINER (vbox2), BORDER);
+      gtk_widget_show (vbox2);
+      gtk_container_add (GTK_CONTAINER (hibernate_button), vbox2);
+
+      icon = xfce_themed_icon_load ("system-hibernate", 32);
+      if (!icon)
+        icon = xfce_themed_icon_load ("xfsm-hibernate", 32);
+      image = gtk_image_new_from_pixbuf (icon);
+      gtk_widget_show (image);
+      gtk_box_pack_start (GTK_BOX (vbox2), image, FALSE, FALSE, 0);
+      g_object_unref (icon);
+
+      label = gtk_label_new (_("Hibernate"));
+      gtk_widget_show (label);
+      gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, FALSE, 0);
+  }
+
+  /* save session */
+  if (!autosave && kiosk_can_save_session)
     {
       checkbox = gtk_check_button_new_with_mnemonic(
           _("_Save session for future logins"));
@@ -328,16 +532,20 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
   /* create small border */
   if (!accessibility)
     xfsm_window_add_border (GTK_WINDOW (dialog));
-	
+  
   /* center dialog on target monitor */
   xfce_gtk_window_center_on_monitor (GTK_WINDOW (dialog), screen, monitor);
 
   /* connect to the shutdown helper */
-  shutdown_helper = xfsm_shutdown_helper_spawn ();
-  if (shutdown_helper == NULL || !kiosk_can_shutdown)
+  if (!kiosk_can_shutdown || 
+      (shutdown_helper = xfsm_shutdown_helper_spawn (NULL)) == NULL)
     {
-      gtk_widget_set_sensitive (radio_reboot, FALSE);
-      gtk_widget_set_sensitive (radio_halt, FALSE);
+      gtk_widget_set_sensitive (reboot_button, FALSE);
+      gtk_widget_set_sensitive (halt_button, FALSE);
+      if (suspend_button)
+        gtk_widget_set_sensitive (suspend_button, FALSE);
+      if (hibernate_button)
+        gtk_widget_set_sensitive (hibernate_button, FALSE);
     }
 
   /* save portion of the root window covered by the dialog */
@@ -348,49 +556,40 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
       gdk_window_raise (dialog->window);
     }
 
-	/* need to realize the dialog first! */
-	gtk_widget_show_now (dialog);
-  gtk_widget_grab_focus (ok_button);
+  /* need to realize the dialog first! */
+  gtk_widget_show_now (dialog);
+  gtk_widget_grab_focus (logout_button);
 
-	/* Grab Keyboard and Mouse pointer */
+  /* Grab Keyboard and Mouse pointer */
   if (!accessibility)
     xfsm_window_grab_input (GTK_WINDOW (dialog));
 
-	/* run the logout dialog */
-	result = gtk_dialog_run (GTK_DIALOG(dialog));
+  /* run the logout dialog */
+  result = gtk_dialog_run (GTK_DIALOG(dialog));
 
-	if (result == GTK_RESPONSE_OK) {
-		*saveSession = autosave || gtk_toggle_button_get_active(
-				GTK_TOGGLE_BUTTON(checkbox));
-    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio_logout)))
-      *shutdownType = SHUTDOWN_LOGOUT;
-    else if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio_reboot)))
-      *shutdownType = SHUTDOWN_REBOOT;
-    else
-      *shutdownType = SHUTDOWN_HALT;
-	}
+  if (result == GTK_RESPONSE_OK) {
+    *saveSession = autosave ? autosave : 
+            gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbox));
+  }
 
   gtk_widget_hide (dialog);
 
   /* ask password */
-  if (result == GTK_RESPONSE_OK && *shutdownType != SHUTDOWN_LOGOUT
+  if (result == GTK_RESPONSE_OK && *shutdownType != XFSM_SHUTDOWN_LOGOUT
       && xfsm_shutdown_helper_need_password (shutdown_helper))
     {
-      if (checkbox != NULL)
-        gtk_widget_destroy (checkbox);
-      gtk_widget_destroy (radio_vbox);
+      gtk_widget_show (ok_button);
+
+      gtk_widget_destroy (vbox);
 
       entry_vbox = gtk_vbox_new (FALSE, BORDER);
-      gtk_box_pack_start (GTK_BOX (vbox), entry_vbox, TRUE, TRUE, 0);
+      gtk_box_pack_start (GTK_BOX (dbox), entry_vbox, TRUE, TRUE, BORDER);
       gtk_widget_show (entry_vbox);
 
-#if GTK_CHECK_VERSION(2,4,0)
-      gtk_image_set_from_stock (GTK_IMAGE (image),
-                                GTK_STOCK_DIALOG_AUTHENTICATION,
-                                GTK_ICON_SIZE_DIALOG);
-#endif
-
-      gtk_label_set_text (GTK_LABEL (label), _("Please enter your password:"));
+      label = gtk_label_new (_("Please enter your password:"));
+      gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
+      gtk_widget_show (label);
+      gtk_box_pack_start (GTK_BOX (entry_vbox), label, FALSE, FALSE, 0);
 
       entry = gtk_entry_new ();
       gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
@@ -417,12 +616,8 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
 
           if (!xfsm_shutdown_helper_send_password (shutdown_helper, pw))
             {
-              gtk_image_set_from_stock (GTK_IMAGE (image),
-                                        GTK_STOCK_DIALOG_ERROR,
-                                        GTK_ICON_SIZE_DIALOG);
-
               gtk_label_set_text (GTK_LABEL (label),
-                                  _("<b>An error occured</b>"));
+                                  _("<b>An error occurred</b>"));
               gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
 
               gtk_container_remove (GTK_CONTAINER (
@@ -461,10 +656,12 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
         }
     }
 
-	gtk_widget_destroy(dialog);
+  gtk_widget_destroy(dialog);
   gtk_widget_destroy(hidden);
 
-	/* Release Keyboard/Mouse pointer grab */
+  shutdown_dialog = NULL;
+
+  /* Release Keyboard/Mouse pointer grab */
   if (!accessibility)
     {
       xfsm_fadeout_destroy (fadeout);
@@ -483,9 +680,8 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
    */
   if (result == GTK_RESPONSE_OK)
     {
-      xfce_rc_set_group (rc, "General");
-      xfce_rc_write_entry (rc, "SessionName", session_name);
-      xfce_rc_write_bool_entry (rc, "SaveOnExit", *saveSession);
+      xfconf_channel_set_string (channel, "/general/SessionName", sessionName);
+      xfconf_channel_set_bool (channel, "/general/SaveOnExit", *saveSession);
     }
   else
     {
@@ -497,40 +693,52 @@ shutdownDialog(gint *shutdownType, gboolean *saveSession)
   if (screenshot_pm != NULL)
     {
       if (result == GTK_RESPONSE_OK)
-        screenshot_save (screenshot_pm, &screenshot_area);
+        screenshot_save (sessionName, screenshot_pm, &screenshot_area);
 
       g_object_unref (G_OBJECT (screenshot_pm));
     }
 #endif
 
-  xfce_rc_close (rc);
-
-	return (result == GTK_RESPONSE_OK);
+  return (result == GTK_RESPONSE_OK);
 }
 
 
 /*
  */
 gint
-xfsm_shutdown(gint type)
+xfsm_shutdown(XfsmShutdownType type)
 {
   gboolean result;
+  GError *error = NULL;
 
-  if (compat_gnome)
-    xfsm_compat_gnome_shutdown ();
+  /* kludge */
+  if (type == XFSM_SHUTDOWN_ASK)
+    {
+      g_warning ("xfsm_shutdown () passed XFSM_SHUTDOWN_ASK.  This is a bug.");
+      type = XFSM_SHUTDOWN_LOGOUT;
+    }
 
-  if (compat_kde)
-    xfsm_compat_kde_shutdown ();
+  /* these two remember if they were started or not */
+  xfsm_compat_gnome_shutdown ();
+  xfsm_compat_kde_shutdown ();
 
   /* kill legacy clients */
   xfsm_legacy_shutdown ();
 
-#ifdef HAVE_SYNC
-  /* sync disk block in-core status with that on disk */
-  sync ();
-#endif
+#if !defined(__NR_ioprio_set) && defined(HAVE_SYNC)
+  /* sync disk block in-core status with that on disk.  if
+   * we have ioprio_set (), then we've already synced. */
+  if (fork () == 0)
+    {
+# ifdef HAVE_SETSID
+      setsid ();
+# endif
+      sync ();
+      _exit (EXIT_SUCCESS);
+    }
+#endif  /* HAVE_SYNC */
 
-  if (type == SHUTDOWN_LOGOUT)
+  if (type == XFSM_SHUTDOWN_LOGOUT)
     return EXIT_SUCCESS;
 
   if (shutdown_helper == NULL)
@@ -539,24 +747,19 @@ xfsm_shutdown(gint type)
       return EXIT_FAILURE;
     }
 
-  if (type == SHUTDOWN_HALT)
-    {
-      result = xfsm_shutdown_helper_send_command (shutdown_helper,
-                                                  XFSM_SHUTDOWN_POWEROFF);
-    }
-  else
-    {
-      result = xfsm_shutdown_helper_send_command (shutdown_helper,
-                                                  XFSM_SHUTDOWN_REBOOT);
-    }
-
+  result = xfsm_shutdown_helper_send_command (shutdown_helper, type, &error);
   xfsm_shutdown_helper_destroy (shutdown_helper);
   shutdown_helper = NULL;
 
   if (!result)
     {
-      /* XXX - graphical feedback ?! */
-      g_warning ("Failed to perform shutdown action!");
+      xfce_message_dialog (NULL, _("Shutdown Failed"),
+                           GTK_STOCK_DIALOG_ERROR,
+                           _("Unable to perform shutdown"),
+                           error->message,
+                           GTK_STOCK_QUIT, GTK_RESPONSE_ACCEPT,
+                           NULL);
+      g_error_free (error);
       return EXIT_FAILURE;
     }
 
